#
#   Code for automatically running tests
#

import glob, os, sys
from os import path
from string import replace, strip

from Errors import PyrexError
import Pyrex
from Pyrex import replace_suffix

if sys.platform == "mac":
  from MacTesting import bbedit_open, mpw_command, run_python_file
  from MacSystem import c_compile
else:
  raise Exception(
    "Unsupported hardware platform: %s" % sys.platform)


class FailureError(Exception):
  pass


failure_flag = 0

def run_tests():
  "Run tests given as command line arguments."
  try:
    try:
      args = sys.argv[1:]
      if args:
        for arg in args:
          run_test(arg)
      else:
        run_all_tests()
    finally:
      if failure_flag:
        print "FAILURES OCCURRED"
      else:
        print "All tests passed."
  except PyrexError, e:
    print repr(e)
    print e
    sys.exit(1)
  
def run_all_tests():
  "Run all tests in the Tests folder of the current dir."
  test_dir = path.join(os.curdir, "Tests")
  run_tests_in_dir(test_dir)

def run_test(item_path):
  "Run a single test or directory of tests."
  try:
    item_name = path.basename(item_path)
    if path.isdir(item_path):
      if item_name[:2] == "R_":
        run_functional_test(item_path)
      else:
        run_tests_in_dir(item_path)
    elif item_name[:2] == "r_":
      run_functional_test(item_path)
    else:
      run_compile_test(item_path)
  except FailureError, e:
    print "TEST FAILED:", e
    ans = raw_input("Continue testing [y/n]? ")
    if ans[:1] <> "y":
      print "Testing aborted."
      sys.exit(1)

def run_tests_in_dir(dir):
  "Run all tests in given directory."
  print "Running tests in", dir
  names = glob.glob(path.join(dir, "*.pyx"))
  for name in names:
    item = path.join(dir, name)
    run_test(item)
  names = os.listdir(dir)
  for name in names:
    if name not in ignore_dir_names:
      item = path.join(dir, name)
      if path.isdir(item):
        run_test(item)

ignore_dir_names = (
  "Reference", "CantTestYet"
)

def run_compile_test(item):
  "Run a single compile-only test."
  print "Running test", item
  dir = path.dirname(item)
  name = path.basename(item)
  is_error_test = (
    name[:2] == "e_" or
    name[:3] == "se_")
  result = Pyrex.compile(item, c_only = 1)
  #print "result =", result.__dict__ ###
  if is_error_test:
    if result.num_errors == 0:
      fail("No errors produced, expected some")
    if result.listing_file is None:
      fail("No listing file produced")
    compare_with_reference(result.listing_file)
  else:
    if result.num_errors <> 0:
      display_files(replace_suffix(item, ".lis"))
      fail("%s errors reported, expected none" %
        result.num_errors)
    if result.c_file is None:
      fail("No C file produced")
    compare_with_reference(result.c_file, show_diffs = 1)
    c_compile(result.c_file)

def run_functional_test_dir(dir, keep_files = 0):
  pyx_files = glob.glob(path.join(dir, "*.pyx"))
  if not pyx_files:
    fail("No .pyx file")
  if len(pyx_files) > 1:
    fail("Too many .pyx files")
  pyx_file = pyx_files[0]
  run_functional_test(pyx_file, keep_files)

def run_functional_test(pyx_file, keep_files = 0):
  "Run a compile, link and execute test."
  result = Pyrex.compile(pyx_file)
  if result.num_errors <> 0:
    fail("Compilation errors")
  new_c = compare_with_reference(result.c_file, show_diffs = 1)
  py_file = replace_suffix(pyx_file, "_t.py")
  out_file = replace_suffix(pyx_file, ".out")
  err_file = replace_suffix(pyx_file, ".err")
  stat = run_python_file(py_file, out_file, err_file)
  if stat:
    fail("Exit status %s" % stat)
  new_output = compare_with_reference(out_file, show_diffs = 0)
  if not keep_files:
    remove_file(replace_suffix(pyx_file, ".lis"))
    if not new_c:
      remove_file(result.c_file)
    remove_file(result.object_file)
    remove_file(result.extension_file)
    remove_file(err_file)

def compare_with_reference(file_in_question, show_diffs):
  dir = path.dirname(file_in_question)
  name = path.basename(file_in_question)
  refdir = path.join(dir, "Reference")
  reference_file = path.join(refdir, name)
  if not path.exists(reference_file):
    print "NEW RESULT:", file_in_question
    #display_files(file_in_question)
    return 1
  lines1 = get_lines(file_in_question)
  lines2 = get_lines(reference_file)
#	len1 = len(lines1)
#	len2 = len(lines2)
#	for i in range(max(len1, len2)):
#		if i >= len1 or i >= len2 or lines1[i] <> lines2[i]:
#			#display_files(reference_file, file_in_question)
#			if show_diffs:
#				display_differences(reference_file, file_in_question)
#			fail("%s differs from reference at line %s" % (
#				name, i))
  if lines1 <> lines2:
    print "%s differs from reference." % name
    show_munged_lines_difference(lines1, lines2)
    ans = raw_input("Show full diffs [y/n]? ")
    if ans[:1] == "y":
      display_differences(reference_file, file_in_question)
    fail("%s differs from reference" % name)
  return 0

def get_lines(filename):
  try:
    f = open(filename)
    lines = f.readlines()
    f.close()
  except IOError, e:
    fail(str(e))
  lines2 = []
  for line in lines:
    line = strip(line)
    if line[:2] == "/*" and line[-2:] == "*/":
      line = ""
    elif line in ignore_lines:
      line = ""
    elif line[-1:] == ":":
      line = line + ";"
    line = replace(line, " ", "")
    # HACK: (void) <-> ()
    line = replace(line, "(void)", "()")
    # HACK: fudge PyNumber_AsLong
    line = replace(line, "PyNumber_AsLong", "PyInt_AsLong")
    # HACK: ignore prototypes
    if line[-9:] == "/*proto*/":
      line = ""
    # HACK: try to ignore old prototypes
    if line[:6] == "static" and line[-1:] == ";":
      line = ""
    elif len(lines2) > 0:
      l1 = lines2[-1]
      l2 = line
      #l1 = replace(l1, "(", "")
      #l1 = replace(l1, ")", "")
      #l2 = replace(l2, "(", "")
      #l2 = replace(l2, ")", "")
      if l1[:-1] == l2[:-1] and \
        l1[-1:] == ";" and l2[-1:] == "{":
          del lines2[-1]
    # HACK: ignore "static"
    if line[:6] == "static":
      line = line[6:]
    # HACK: ignore module init func onwards
    if line[:8] == "voidinit" and \
        line[-1:] in ("{", ";") and \
        (line[-7:-1] == "(void)" or line[-3:-1] == "()"):
      #print "Stopping at", line ###
      break
    # END HACKS
    if line:
      lines2.append(line)
  return lines2

ignore_lines = (
  """static PyObject *__pyx_b;""",
  """__pyx_b = PyImport_AddModule("__builtin__");""",
  """PyDict_SetItemString(__pyx_d, "__builtins__", __pyx_b);""",
  """static PyObject *__Pyx_UnpackItem(PyObject *, int);""",
  """static int __Pyx_EndUnpack(PyObject *, int);""",
  """static int __Pyx_PrintItem(PyObject *);""",
  """static int __Pyx_PrintNewline(void);""",
  """static void __Pyx_ReRaise(void);""",
  """static void __Pyx_RaiseWithTraceback(PyObject *, PyObject *, PyObject *);""",
  """static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list);""",
  """static PyObject *__Pyx_Import(PyObject *name, PyObject *globals, PyObject *from_list);""",
  """static PyObject *__Pyx_GetExcValue(void);""",
  """static PyObject *__Pyx_GetName(PyObject *dict, char *name);""",
)

def show_munged_lines_difference(lines1, lines2):
  print "Flexidiff:"
  for i in range(min(len(lines1), len(lines2))):
    if lines1[i] <> lines2[i]:
      print "%4d< %s" % (i, repr(lines1[i]))
      print "%4d> %s" % (i, repr(lines2[i]))
      break
  if len(lines1) > len(lines2):
    print "%4d< %s" % (len(lines2), repr(lines1[len(lines2)]))
  if len(lines2) > len(lines1):
    print "%4d> %s" % (len(lines1), repr(lines2[len(lines1)]))

def fail(mess):
  global failure_flag
  failure_flag = 1
  raise FailureError(mess)

def display_files(*filenames):
  if sys.platform == "mac":
    for filename in filenames:
      bbedit_open(filename)

def display_differences(file1, file2):
  if sys.platform == "mac":
    mpw_command("CompareFiles %s %s" % (file1, file2))

def remove_file(file):
  print "Removing:", file ###
  if file:
    try:
      os.unlink(file)
    except IOError:
      pass
